@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 from __future__ import unicode_literals
+
 from django.conf import settings
 if 'django_select2' in settings.INSTALLED_APPS:
     from django_select2.fields import AutoModelSelect2Field
@@ -7,9 +8,18 @@ else:
     raise ImportError('django_select2 not configured')
 
 
-class PageSearchField(AutoModelSelect2Field):
+class LinkSearchField(AutoModelSelect2Field):
     empty_value = []
-    search_fields = ['title_set__title__icontains', 'title_set__menu_title__icontains', 'title_set__slug__icontains']
+
+    class Media:
+        js = (settings.STATIC_ROOT + 'node_modules/jquery/dist/jquery.min.js',)
+
+    def __init__(self, *args, **kwargs):
+        try:
+            self.search_fields = kwargs.pop('search_fields')
+        except KeyError:
+            pass
+        super(LinkSearchField, self).__init__(*args, **kwargs)
 
     def security_check(self, request, *args, **kwargs):
         user = request.user
@@ -20,4 +30,4 @@ class PageSearchField(AutoModelSelect2Field):
     def prepare_value(self, value):
         if not value:
             return None
-        return super(PageSearchField, self).prepare_value(value)
+        return super(LinkSearchField, self).prepare_value(value)
