@@ -1,17 +1,43 @@
 # -*- coding: utf-8 -*-
 from __future__ import unicode_literals
+
+from django.conf import settings
 from django.contrib.sites.models import Site
-from django.db.models import get_model
-from django.forms import fields
-from django.forms.widgets import TextInput
-from django.forms.models import ModelForm
-from django.utils.translation import ugettext_lazy as _
 from django.core.exceptions import ObjectDoesNotExist
+from django.apps import apps
+from django.forms import fields
+from django.forms.models import ModelForm
+from django.utils.module_loading import import_string
+from django.utils.translation import ugettext_lazy as _
 from cms.models import Page
-try:
-    from .fields import PageSearchField as PageSelectFormField
-except ImportError:
-    from cms.forms.fields import PageSelectFormField
+from cmsplugin_cascade.models import CascadePage
+from cmsplugin_cascade.utils import validate_link
+
+if 'django_select2' in settings.INSTALLED_APPS:
+    Select2Widget = import_string('django_select2.forms.Select2Widget')
+else:
+    Select2Widget = import_string('django.forms.widgets.Select')
+
+
+class SelectWidget(Select2Widget):
+    @property
+    def media(self):
+        parent_media = super(SelectWidget, self).media
+        # prepend JS snippet to re-add 'jQuery' to the global namespace
+        parent_media._js.insert(0, 'cascade/js/admin/jquery.restore.js')
+        return parent_media
+
+
+class LinkSearchField(fields.ChoiceField):
+    def __init__(self, *args, **kwargs):
+        kwargs.setdefault('widget', SelectWidget)
+        super(LinkSearchField, self).__init__(*args, **kwargs)
+
+    def clean(self, value):
+        try:
+            return int(value)
+        except (TypeError, ValueError):
+            pass
 
 
 class LinkForm(ModelForm):
@@ -20,40 +46,66 @@ class LinkForm(ModelForm):
     populate the glossary of the model.
     """
     LINK_TYPE_CHOICES = (('cmspage', _("CMS Page")), ('exturl', _("External URL")), ('email', _("Mail To")),)
-    link_type = fields.ChoiceField()
-    cms_page = PageSelectFormField(required=False, label='',
+    link_type = fields.ChoiceField(label=_("Link"), help_text=_("Type of link"))
+    cms_page = LinkSearchField(required=False, label='',
         help_text=_("An internal link onto CMS pages of this site"))
+    section = fields.ChoiceField(required=False, label='',
+        help_text=_("Page bookmark"))
     ext_url = fields.URLField(required=False, label='', help_text=_("Link onto external page"))
     mail_to = fields.EmailField(required=False, label='', help_text=_("Open Email program with this address"))
 
     class Meta:
         fields = ('glossary',)
 
-    def __init__(self, raw_data=None, *args, **kwargs):
+    def __init__(self, data=None, *args, **kwargs):
         instance = kwargs.get('instance')
-        initial = instance and dict(instance.glossary) or {'link': {'type': 'cmspage'}}
+        default_link_type = {'type': self.LINK_TYPE_CHOICES[0][0]}
+        initial = dict(instance.glossary) if instance else {'link': default_link_type}
         initial.update(kwargs.pop('initial', {}))
+        initial.setdefault('link', {'type': default_link_type})
         link_type = initial['link']['type']
         self.base_fields['link_type'].choices = self.LINK_TYPE_CHOICES
         self.base_fields['link_type'].initial = link_type
-        if raw_data and raw_data.get('shared_glossary'):
+        if data and data.get('shared_glossary'):
             # convert this into an optional field since it is disabled with ``shared_glossary`` set
             self.base_fields['link_type'].required = False
+        set_initial_linktype = getattr(self, 'set_initial_{}'.format(link_type), None)
+
+        # populate Select field for choosing a CMS page
         try:
             site = instance.page.site
         except AttributeError:
             site = Site.objects.get_current()
-        self.base_fields['cms_page'].queryset = Page.objects.drafts().on_site(site)
-        set_initial_linktype = getattr(self, 'set_initial_{0}'.format(link_type), None)
+        choices = ((p.pk, '{0} ({1})'.format(p.get_page_title(), p.get_absolute_url()))
+                   for p in Page.objects.drafts().on_site(site))
+        self.base_fields['cms_page'].choices = choices
+
         if callable(set_initial_linktype):
             set_initial_linktype(initial)
-        kwargs.update(initial=initial)
-        super(LinkForm, self).__init__(raw_data, *args, **kwargs)
+        self._preset_section(data, initial)
+        super(LinkForm, self).__init__(data, initial=initial, *args, **kwargs)
+
+    def _preset_section(self, data, initial):
+        choices = [(None, _("Page root"))]
+        try:
+            if data:
+                cms_page = Page.objects.get(pk=data['cms_page'])
+            else:
+                cms_page = Page.objects.get(pk=initial['link']['pk'])
+        except (KeyError, ValueError, ObjectDoesNotExist):
+            pass
+        else:
+            CascadePage.assure_relation(cms_page)
+            for key, val in cms_page.cascadepage.glossary.get('element_ids', {}).items():
+                choices.append((key, val))
+
+        self.base_fields['section'].initial = initial['link'].get('section')
+        self.base_fields['section'].choices = choices
 
     def clean_glossary(self):
         """
-        This method rectifies the behavior of JSONFormFieldBase.clean which
-        converts the value of empty fields to None, while it shall be an empty dict.
+        This method rectifies the behavior of JSONFormFieldBase.clean which converts
+        the value of empty fields to None, although it shall be an empty dict.
         """
         glossary = self.cleaned_data['glossary']
         if glossary is None:
@@ -66,6 +118,8 @@ class LinkForm(ModelForm):
             if 'link_data' in cleaned_data:
                 cleaned_data['glossary'].update(link=cleaned_data['link_data'])
                 del self.cleaned_data['link_data']
+            elif 'link_type' in cleaned_data:
+                cleaned_data['glossary'].update(link={'type': cleaned_data['link_type']})
             else:
                 cleaned_data['glossary'].update(link={'type': 'none'})
         return cleaned_data
@@ -75,8 +129,13 @@ class LinkForm(ModelForm):
             self.cleaned_data['link_data'] = {
                 'type': 'cmspage',
                 'model': 'cms.Page',
-                'pk': self.cleaned_data['cms_page'] and self.cleaned_data['cms_page'].pk or None
+                'pk': self.cleaned_data['cms_page'],
             }
+            validate_link(self.cleaned_data['link_data'])
+
+    def clean_section(self):
+        if self.cleaned_data.get('link_type') == 'cmspage':
+            self.cleaned_data['link_data']['section'] = self.cleaned_data['section']
 
     def clean_ext_url(self):
         if self.cleaned_data.get('link_type') == 'exturl':
@@ -91,8 +150,9 @@ class LinkForm(ModelForm):
 
     def set_initial_cmspage(self, initial):
         try:
-            Model = get_model(*initial['link']['model'].split('.'))
-            initial['cms_page'] = Model.objects.get(pk=initial['link']['pk'])
+            # check if that page still exists, otherwise return nothing
+            Model = apps.get_model(*initial['link']['model'].split('.'))
+            initial['cms_page'] = Model.objects.get(pk=initial['link']['pk']).pk
         except (KeyError, ObjectDoesNotExist):
             pass
 
@@ -108,6 +168,13 @@ class LinkForm(ModelForm):
         except KeyError:
             pass
 
+    @classmethod
+    def get_form_class(cls):
+        """
+        Hook to return a form class for editing a CMSPlugin inheriting from ``LinkPluginBase``.
+        """
+        return cls
+
     @classmethod
     def unset_required_for(cls, sharable_fields):
         """
@@ -118,21 +185,12 @@ class LinkForm(ModelForm):
             cls.base_fields['link_type'].required = False
 
 
-class TextLinkForm(LinkForm):
+class TextLinkFormMixin(object):
     """
-    Form class with the additional fake field ``link_content`` used to store the content of pure
-    text links.
+    To be used in combination with `LinkForm` for easily accessing the field `link_content`.
     """
-    link_content = fields.CharField(required=False, label=_("Link Content"),
-        # replace auto-generated id so that CKEditor automatically transfers the text into this input field
-        widget=TextInput(attrs={'id': 'id_name'}), help_text=_("Content of Link"))
-
     def clean(self):
-        """
-        link_content intentionally was rendered outside the glossary field, now move this content
-        back to the ``glossary``.
-        """
-        cleaned_data = super(TextLinkForm, self).clean()
+        cleaned_data = super(TextLinkFormMixin, self).clean()
         if self.is_valid():
             cleaned_data['glossary'].update(link_content=cleaned_data['link_content'])
         return cleaned_data
