@@ -1,32 +1,38 @@
 # -*- coding: utf-8 -*-
 from __future__ import unicode_literals
+
 import re
 import json
-from django import VERSION as DJANGO_VERSION
+try:
+    from html.parser import HTMLParser  # py3
+except ImportError:
+    from HTMLParser import HTMLParser  # py2
 from django.core.exceptions import ValidationError
 from django.forms import widgets
 from django.utils import six
 from django.utils.safestring import mark_safe
 from django.utils.html import escape, format_html, format_html_join
 from django.utils.translation import ugettext_lazy as _, ugettext
-from .fields import PartialFormField
+from .fields import GlossaryField
 
 
 class JSONMultiWidget(widgets.MultiWidget):
     """Base class for MultiWidgets using a JSON field in database"""
-    def __init__(self, partial_fields):
-        self.partial_fields = partial_fields[:]
+    html_parser = HTMLParser()
+
+    def __init__(self, glossary_fields):
+        self.glossary_fields = list(glossary_fields)
         self.normalized_fields = []
-        for field in self.partial_fields:
-            if isinstance(field, PartialFormField):
+        for field in self.glossary_fields:
+            if isinstance(field, GlossaryField):
                 self.normalized_fields.append(field)
             elif isinstance(field, (list, tuple)):
-                self.normalized_fields.extend([f for f in field if isinstance(f, PartialFormField)])
+                self.normalized_fields.extend([f for f in field if isinstance(f, GlossaryField)])
             else:
-                raise ValueError('Given fields must be of type PartialFormField or list of thereof')
+                raise ValueError("Given fields must be of type GlossaryField or list of thereof")
         unique_keys = set([field.name for field in self.normalized_fields])
         if len(self.normalized_fields) > len(unique_keys):
-            raise ValueError('List of partial_fields may contain only unique keys')
+            raise ValueError('List of glossary_fields may contain only unique keys')
         super(JSONMultiWidget, self).__init__((field.widget for field in self.normalized_fields))
 
     def decompress(self, values):
@@ -54,16 +60,19 @@ class JSONMultiWidget(widgets.MultiWidget):
         values = self.decompress(values)
         field_attrs = dict(**attrs)
         render_fieldsets = []
-        for fieldset in self.partial_fields:
+        for fieldset in self.glossary_fields:
             render_fields = []
             if not isinstance(fieldset, (list, tuple)):
                 fieldset = [fieldset]
             for field in fieldset:
                 field_attrs['id'] = '{id}_{0}'.format(field.name, **attrs)
+                field_value = values.get(field.name)
+                if isinstance(field_value, six.string_types):
+                    field_value = self.html_parser.unescape(field_value)
                 render_fields.append((
                     field.name,
                     six.text_type(field.label),
-                    field.widget.render(field.name, values.get(field.name), field_attrs),
+                    field.widget.render(field.name, field_value, field_attrs),
                     six.text_type(field.help_text),
                 ))
             html = format_html_join('',
@@ -73,14 +82,8 @@ class JSONMultiWidget(widgets.MultiWidget):
         return format_html_join('\n', '<div class="glossary-widget">{0}</div>', render_fieldsets)
 
 
-if DJANGO_VERSION[:2] < (1, 6):
-    input_widget = widgets.TextInput
-else:
-    input_widget = widgets.NumberInput
-
-
-class NumberInputWidget(input_widget):
-    validation_pattern = re.compile('^\d+$')
+class NumberInputWidget(widgets.NumberInput):
+    validation_pattern = re.compile('^-?\d+$')
     required = True
     required_message = _("In '%(label)s': This field is required.")
     invalid_message = _("In '%(label)s': Value '%(value)s' shall contain a valid number.")
@@ -90,33 +93,43 @@ class NumberInputWidget(input_widget):
             raise ValidationError(self.validation_message, code='invalid', params={'value': value})
 
 
-def _compile_validation_pattern(widget, units):
-    """
-    Assure that passed in units are valid size units.
-    Return a tuple with a regular expression to be used for validating and a message if this
-    validation failed.
-    """
-    for u in units:
-        if u not in widget.POSSIBLE_UNITS:
-            raise ValidationError('{0} is not a valid unit for CascadingSizeField'.format(u))
-    endings = (' %s ' % ugettext("or")).join("'%s'" % u.replace('%', '%%') for u in units)
-    params = {'label': '%(label)s', 'value': '%(value)s', 'field': '%(field)s', 'endings': endings}
-    return re.compile(r'^(\d+)\s*({0})$'.format('|'.join(units))), widget.invalid_message % params
+class CascadingSizeWidgetMixin(object):
+    POSSIBLE_UNITS = ('px', 'em', '%')
+    required_message = _("In '%(label)s': This field is required.")
+    invalid_message = _("In '%(label)s': Value '%(value)s' shall contain a valid number, ending in %(endings)s.")
+
+    def compile_validation_pattern(self, units=None):
+        """
+        Assure that passed in units are valid size units, or if missing, use all possible units.
+        Return a tuple with a regular expression to be used for validating and an error message
+        in case this validation failed.
+        """
+        if units is None:
+            units = list(self.POSSIBLE_UNITS)
+        else:
+            for u in units:
+                if u not in self.POSSIBLE_UNITS:
+                    raise ValidationError('{} is not a valid unit for a size field'.format(u))
+        regex = re.compile(r'^(-?\d+)({})$'.format('|'.join(units)))
+        endings = (' %s ' % ugettext("or")).join("'%s'" % u.replace('%', '%%') for u in units)
+        params = {'label': '%(label)s', 'value': '%(value)s', 'field': '%(field)s', 'endings': endings}
+        return regex, self.invalid_message % params
 
 
-class CascadingSizeWidget(widgets.TextInput):
+class CascadingSizeWidget(CascadingSizeWidgetMixin, widgets.TextInput):
     """
     Use this field for validating Input Fields containing a value ending in ``px``, ``em`` or ``%``.
     Use it for values representing a margin, padding, width or height.
     """
-    POSSIBLE_UNITS = ['px', 'em', '%']
     DEFAULT_ATTRS = {'style': 'width: 5em;'}
-    required_message = _("In '%(label)s': This field is required.")
-    invalid_message = _("In '%(label)s': Value '%(value)s' shall contain a valid number, ending in %(endings)s.")
 
-    def __init__(self, allowed_units=POSSIBLE_UNITS, attrs=DEFAULT_ATTRS, required=True):
-        self.validation_pattern, self.invalid_message = _compile_validation_pattern(self, allowed_units)
-        self.required = required
+    def __init__(self, allowed_units=None, attrs=None, required=None):
+        if required is not None:
+            self.required = required
+        if attrs is None:
+            attrs = self.DEFAULT_ATTRS
+        self.validation_pattern, self.invalid_message = self.compile_validation_pattern(
+            units=allowed_units)
         super(CascadingSizeWidget, self).__init__(attrs=attrs)
 
     def validate(self, value):
@@ -133,7 +146,7 @@ class CascadingSizeWidget(widgets.TextInput):
 class ColorPickerWidget(widgets.MultiWidget):
     """
     Use this field to enter a color value. Clicking onto this widget will pop up a color picker.
-    The value passed to the PartialField is guaranteed to be in #rgb format.
+    The value passed to the GlossaryField is guaranteed to be in #rgb format.
     """
     DEFAULT_COLOR = '#ffffff'
     DEFAULT_ATTRS = {'style': 'width: 5em;', 'type': 'color'}
@@ -246,13 +259,79 @@ class MultipleTextInputWidget(widgets.MultiWidget):
                 raise ValidationError(self.invalid_message, code='invalid', params=params)
 
 
-class MultipleCascadingSizeWidget(MultipleTextInputWidget):
-    POSSIBLE_UNITS = ['px', 'em', '%']
+class MultipleCascadingSizeWidget(CascadingSizeWidgetMixin, MultipleTextInputWidget):
     DEFAULT_ATTRS = {'style': 'width: 4em;'}
-    required_message = _("In '%(label)s': Field '%(field)s' is required.")
     invalid_message = _("In '%(label)s': Value '%(value)s' for field '%(field)s' shall contain a valid number, ending in %(endings)s.")
 
-    def __init__(self, labels, allowed_units=POSSIBLE_UNITS, attrs=DEFAULT_ATTRS, required=True):
-        self.validation_pattern, self.invalid_message = _compile_validation_pattern(self, allowed_units)
-        self.required = required
-        super(MultipleCascadingSizeWidget, self).__init__(labels, attrs=attrs)
+    def __init__(self, labels, allowed_units=None, attrs=None, required=True):
+        if attrs is None:
+            attrs = self.DEFAULT_ATTRS
+        self.validation_pattern, self.invalid_message = self.compile_validation_pattern(
+            units=allowed_units)
+        super(MultipleCascadingSizeWidget, self).__init__(labels, required=required, attrs=attrs)
+
+
+class SetBorderWidget(widgets.MultiWidget):
+    """
+    Use this field to enter the three values of a border: width style color.
+    """
+    DEFAULT_COLOR = '#000000'
+    BORDER_STYLES = ['dashed', 'dotted', 'double', 'groove', 'hidden', 'inset', 'none', 'outset',
+                     'ridge', 'solid']
+
+    invalid_border_message = _("In '%(label)s': Value '%(value)s' is not a valid border style.")
+    color_validation_pattern = re.compile('^#[0-9a-f]{3}([0-9a-f]{3})?$')
+    invalid_color_message = _("In '%(label)s': Value '%(value)s' is not a valid color.")
+
+    def __init__(self, attrs=None):
+        widget_list = [
+            CascadingSizeWidget(),
+            widgets.Select(choices=[(s, s) for s in self.BORDER_STYLES]),
+            widgets.TextInput(attrs={'style': 'width: 5em;', 'type': 'color'})
+        ]
+        super(SetBorderWidget, self).__init__(widget_list)
+
+    def __iter__(self):
+        for name in ('width', 'style', 'color'):
+            yield name
+
+    def decompress(self, values):
+        if not isinstance(values, (list, tuple)) or len(values) != 3:
+            values = ('0px', 'none', self.DEFAULT_COLOR,)
+        return values
+
+    def value_from_datadict(self, data, files, name):
+        values = (
+            escape(data.get('{0}-width'.format(name), '0px')),
+            escape(data.get('{0}-style'.format(name), 'none')),
+            escape(data.get('{0}-color'.format(name), self.DEFAULT_COLOR)),
+        )
+        return values
+
+    def render(self, name, values, attrs):
+        width, style, color = values
+        elem_id = attrs['id']
+        attrs = dict(attrs)
+        html = '<div class="clearfix">'
+        key, attrs['id'] = '{0}-width'.format(name), '{0}_width'.format(elem_id)
+        html += format_html('<div class="sibling-field">{0}</div>', self.widgets[0].render(key, width, attrs))
+        key, attrs['id'] = '{0}-style'.format(name), '{0}_style'.format(elem_id)
+        html += format_html('<div class="sibling-field">{0}</div>', self.widgets[1].render(key, style, attrs))
+        key, attrs['id'] = '{0}-color'.format(name), '{0}_color'.format(elem_id)
+        html += format_html('<div class="sibling-field">{0}</div>', self.widgets[2].render(key, color, attrs))
+        html += '</div>'
+        return mark_safe(html)
+
+    def validate(self, values, key):
+        if key == 'width':
+            self.widgets[0].validate(values[0])
+        elif key == 'style':
+            style = values[1]
+            if style not in self.BORDER_STYLES:
+                raise ValidationError(self.invalid_border_message, code='invalid', params={'value': style})
+        elif key == 'color':
+            color = values[2]
+            if not self.color_validation_pattern.match(color):
+                raise ValidationError(self.invalid_color_message, code='invalid', params={'value': color})
+        else:
+            raise KeyError("{} is not a valid border widget attribute.".format(key))
